class ResourceMask:
	def __init__(self, read: bool, write: bool, flash: bool) -> None:
		self.read = read
		self.write = write
		self.flash = flash
	
	def getInteger(self) -> int:
		output = 0
		if self.write == True:
			output += 1
		if self.read == True:
			output += 2
		if self.flash == True:
			output += 64
		return output

	def setFromInteger(self, n: int) -> None:
		self.write = (n & 1) != 0
		self.read = (n & 2) != 0
		self.flash = (n & 64) != 0



from canlib import canlib, Frame
from typing import List, Tuple
from winnie import listops
from winnie import resourceMask as rm
from winnie import formatting
from winnie import verification
 
class Connection:
	def __init__(self, channel: canlib.Channel, id: int, debug: bool = False):
		self.connected = False
		self.channel = channel
		self.counter = 0
		self.id = id
		self.debug = debug
	
	def sendFrame(self, frame: Frame) -> bytearray:
		self.channel.write(frame)
		self.channel.writeSync(timeout=500)
		result = self.channel.read(timeout=500)
		return result.data

	def sendMessage(self, message: bytearray) -> bytearray:
		if self.debug == True:
			formatting.printByteArrayWithLabel("Message: ", message)
		if self.connected == False and message[0] != 0x01:
			raise RuntimeError("Connection must be established before sending a message")
		verification.verifyMessage(message)
		# Construct and send the frame
		frame = Frame(id_=self.id, data=message)
		result = self.sendFrame(frame)
		if self.debug == True:
			formatting.printByteArrayWithLabel("Response: ", result)

		currentCounter = self.counter
		self.counter += 0x01
		verification.verifyResponse(result, currentCounter)

		return result, currentCounter

	def connect(self, stationID: int) -> bool:
		if self.debug == True:
			print("CONNECT")

		splitID = listops.splitNumberByBytes(stationID, bigEndian=False)
		message = bytearray([0x01, self.counter, splitID[0], splitID[1], 0, 0, 0, 0])
		response, msgCounter = self.sendMessage(message)
		if self.checkForAcknowledgement(response) == True:
			self.connected = True
			return True
		else:
			raise RuntimeError("Connection failed")
	
	def disconnect(self, stationID: int, temporary=False) -> bool:
		if self.debug == True:
			print("DISCONNECT")

		splitID = listops.splitNumberByBytes(stationID, bigEndian=False)
		temporaryByte = 0x01
		if temporary == True:
			temporaryByte = 0x00
		message = bytearray([0x07, self.counter, temporaryByte, 0, splitID[0], splitID[1], 0, 0])
		response, msgCounter = self.sendMessage(message)
		if self.checkForAcknowledgement(response) == True:
			self.connected = False
			return True
		else:
			raise RuntimeError("Disconnect failed")
	
	def exchangeID(self) -> Tuple[rm.ResourceMask, rm.ResourceMask]:
		if self.debug == True:
			print("EXCHANGE_ID")

		message = bytearray([0x17, self.counter, 0, 0, 0, 0, 0, 0])
		response, msgCounter = self.sendMessage(message)
		# Initialise two resource mask objects
		availabilityMask = rm.ResourceMask(False, False, False)
		protectionMask = rm.ResourceMask(False, False, False)
		# Put in the data from the response
		availabilityMask.setFromInteger(response[5])
		protectionMask.setFromInteger(response[6])
		return availabilityMask, protectionMask

	def getSeed(self, resourceMask: rm.ResourceMask) -> List[int]:
		if self.debug == True:
			print("GET_SEED")

		message = bytearray([0x12, self.counter, resourceMask.getInteger(), 0, 0, 0, 0, 0])
		response, msgCounter = self.sendMessage(message)
		return response[4:]

	def unlock(self, key: bytearray) -> rm.ResourceMask:
		if self.debug == True:
			print("UNLOCK")
		if len(key) != 6:
			raise ValueError(f"Key must be 6 bytes long, was {len(key)} bytes long")
		message = bytearray([0x13, self.counter])
		message.extend(key)
		response, msgCounter = self.sendMessage(message)
		result = rm.ResourceMask(False, False, False)
		result.setFromInteger(response[3])
		return result

	def setMemoryTransferAddress(self, mtaNumber: int, extension: int, address: int) -> bool:
		if self.debug == True:
			print("SET_MTA")

		if mtaNumber != 0 and mtaNumber != 1:
			raise ValueError("Memory transfer address number must be 0 or 1")
		# Construct the message
		message = [0x02, self.counter, mtaNumber, extension]
		message.extend(listops.splitNumberByBytes(address, bigEndian=True))
		# Add leading 0s to address if necessary (this is why we need to switch to byte arrays)
		if len(message) < 8:
			zerosToAdd = 8 - len(message)
			message = message[:4] + [0] * zerosToAdd + message[4:]
		# Send message and handle response
		message = bytearray(message)
		response, msgCounter = self.sendMessage(message)
		return True

	def upload(self, blockSize: int) -> bytearray:
		if self.debug == True:
			print("UPLOAD")

		if blockSize > 5:
			raise ValueError("Block size must be 5 bytes or less")
		message = bytearray([0x04, self.counter, blockSize, 0, 0, 0, 0, 0])
		response, msgCounter = self.sendMessage(message)
		return response[3:3+blockSize]
	
	def getCCPVersion(self, mainVersion: int, release: int) -> Tuple[int, int]:
		if self.debug == True:
			print("GET_CCP_VERSION")

		message = bytearray([0x1B, self.counter, mainVersion, release, 0, 0, 0, 0])
		response, msgCounter = self.sendMessage(message)
		returnedMainVersion = int(response[3])
		returnedRelease = int(response[4])
		return returnedMainVersion, returnedRelease

	def download(self, data: bytearray) -> Tuple[int, int]:
		if self.debug == True:
			print("DOWNLOAD")

		dataLength = len(data)
		if dataLength > 5:
			raise ValueError("Data must be 5 bytes or less")
		message = bytearray([0x03, self.counter, dataLength])
		message.extend(data)
		message = listops.padToLength(message, 8, padding=0)
		response, msgCounter = self.sendMessage(message)
		newExtension = int(response[3])
		newAddress = listops.listToInt(list(response[4:8]))
		return newExtension, newAddress



commandCodes = {
    0x01: "CONNECT",
    0x1B: "GET_CCP_VERSION",
    0x17: "EXCHANGE_ID",
    0x12: "GET_SEED",
    0x13: "UNLOCK",
    0x02: "SET_MTA",
    0x03: "DNLOAD",
    0x23: "DNLOAD_6",
    0x04: "UPLOAD",
    0x0F: "SHORT_UP",
    0x11: "SELECT_CAL_PAGE",
    0x14: "GET_DAQ_SIZE",
    0x15: "SET_DAQ_POINTER",
    0x16: "WRITE_DAQ",
    0x06: "START_STOP",
    0x07: "DISCONNECT",
    0x0C: "SET_S_STATUS",
    0x0D: "GET_S_STATUS",
    0x0E: "BUILD_CHKSUM",
    0x10: "CLEAR_MEMORY",
    0x18: "PROGRAM",
    0x22: "PROGRAM_6",
    0x19: "MOVE",
    0x05: "TEST",
    0x09: "GET_ACTIVE_CAL_PAGE",
    0x08: "START_STOP_ALL",
    0x20: "DIAG_SERVICE",
    0x21: "ACTION_SERVICE"
}



from typing import List

def printHexList(text: str, n: List[int]):
	result = [f"0x{num:02X}" for num in n]
	print(text + str(result))

def printByteArrayWithLabel(text: str, n: bytearray):
	intList = list(n)
	printHexList(text, intList)



from winnie import constants

def verifyMessage(message: bytearray) -> bool:
	if len(message) != 8:
		raise ValueError("Messages must be 8 bytes long")
	if message[0] not in constants.commandCodes:
		raise ValueError(f"{message[0]} is not a valid command code")
	return True

def verifyReceivedCounter(response: bytearray, sentCounter: int) -> bool:
	if response[0] != 0xFF:
		raise ValueError("Cannot verify counter of messages other than command return messages")
	if response[2] != sentCounter:
		raise RuntimeError(f"Received counter ({response[2]}) does not match sent counter ({sentCounter})")
	return True

def checkForAcknowledgement(message: bytearray) -> bool:
	if message[0] == 0xFF or message[0] == 0xFE:
		if message[1] == 0x00:
			return True
		else:
			return False
	raise ValueError("Cannot check for acknowledgement of data acquisition message")

def verifyResponse(response: bytearray, sentCounter: int) -> bool:
	packetID = response[0]
	# Command return message
	if packetID == 0xFF:
		verifyReceivedCounter(response, sentCounter)
		checkForAcknowledgement(response)
	# Event message
	elif packetID == 0xFE:
		checkForAcknowledgement(response)
	# Data Acquisition Message
	else:		
		pass
	return True



from typing import List

def splitNumberByBytes(num: int, bigEndian: bool = True) -> List[int]:
	output = []
	while num > 0:
		output.append(num % 256)
		# Shift the number 8 bits to the right
		num //= 256
	if bigEndian == True:
		output.reverse()
	return output

def padToLength(l: List, targetLength: int, padding=None) -> List:
	currentLength = len(l)
	if currentLength > targetLength:
		raise ValueError(f"List too long ({currentLength}) for target length ({targetLength})")
	l.extend([padding] * targetLength - currentLength)
	return l

def listToInt(l: List[int]) -> int:
	result = 0
	for i in l:
		result = (result << 8) | i
	return result
